﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Programmable {
  [Item("Programmable Problem (multi-objective)", "Represents a multi-objective problem that can be programmed.")]
  [Creatable("Problems")]
  [StorableClass]
  public class MultiObjectiveProgrammableProblem : MultiObjectiveHeuristicOptimizationProblem<IMultiObjectiveProgrammableProblemEvaluator, ISolutionCreator>, IParameterizedNamedItem, IStorableContent {
    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return Common.Resources.VSImageLibrary.Script; }
    }

    public new ParameterCollection Parameters {
      get { return base.Parameters; }
    }
    IKeyedItemCollection<string, IParameter> IParameterizedItem.Parameters {
      get { return Parameters; }
    }

    public IValueParameter<IMultiObjectiveProblemDefinition> ProblemDefinitionParameter {
      get { return (IValueParameter<IMultiObjectiveProblemDefinition>)Parameters["ProblemDefinition"]; }
    }

    protected IValueParameter<IEncoding> EncodingParameter {
      get { return (IValueParameter<IEncoding>)Parameters["Encoding"]; }
    }

    public IMultiObjectiveProblemDefinition ProblemDefinition {
      get { return ProblemDefinitionParameter.Value; }
      set { ProblemDefinitionParameter.Value = value; }
    }

    [Storable]
    protected List<IParameter> DynamicEncodingParameters;

    [StorableConstructor]
    protected MultiObjectiveProgrammableProblem(bool deserializing) : base(deserializing) { }
    protected MultiObjectiveProgrammableProblem(MultiObjectiveProgrammableProblem original, Cloner cloner)
      : base(original, cloner) {
      DynamicEncodingParameters = original.DynamicEncodingParameters.Select(cloner.Clone).ToList();
      RegisterEventHandlers();
    }
    public MultiObjectiveProgrammableProblem()
      : base(new MultiObjectiveEvaluator(), new MultiEncodingCreator()) {
      Parameters.Add(new ValueParameter<IMultiObjectiveProblemDefinition>("ProblemDefinition", "Defines the problem.", new MultiObjectiveProblemScript() { Name = Name }));
      Parameters.Add(new ValueParameter<IEncoding>("Encoding", "Describes which parameters exist, what they're called, what type they are and their bounds if any."));

      DynamicEncodingParameters = new List<IParameter>();

      Operators.Add(new MultiObjectiveAnalyzer());
      Operators.Add(Evaluator);

      RegisterEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MultiObjectiveProgrammableProblem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable UnusedMember.Local
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }
    // ReSharper restore UnusedMember.Local

    private void RegisterEventHandlers() {
      ProblemDefinitionParameter.ValueChanged += ProblemDefinitionParameterOnValueChanged;
      RegisterHostInstanceChanges();
    }

    private void ProblemDefinitionParameterOnValueChanged(object sender, EventArgs eventArgs) {
      RegisterHostInstanceChanges();
      Parameterize();
    }

    private void RegisterHostInstanceChanges() {
      ProblemDefinitionParameter.Value.ProblemDefinitionChanged += ProblemDefinitionChanged;
      ProblemDefinitionParameter.Value.NameChanged += ProblemDefinitionHostOnNameChanged;
    }

    private void ProblemDefinitionHostOnNameChanged(object sender, EventArgs eventArgs) {
      if (sender != ProblemDefinitionParameter.Value) return;
      Name = ProblemDefinitionParameter.Value.Name;
    }

    protected override void OnNameChanged() {
      base.OnNameChanged();
      ProblemDefinitionParameter.Value.Name = Name;
    }

    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      Parameterize();
    }

    protected virtual void ProblemDefinitionChanged(object sender, EventArgs eventArgs) {
      Parameterize();
    }

    protected virtual void Parameterize() {
      var definition = ProblemDefinitionParameter.Value;
      if (definition == null) return;

      IEncoding encoding = definition.Encoding;
      EncodingParameter.Value = encoding;
      Maximization = new BoolArray(definition.Maximization);

      Evaluator.EncodingParameter.ActualName = EncodingParameter.Name;
      Evaluator.ProblemDefinitionParameter.ActualName = ProblemDefinitionParameter.Name;
      foreach (var evalOp in Operators.OfType<IMultiObjectiveProgrammableProblemEvaluator>()) {
        evalOp.EncodingParameter.ActualName = EncodingParameter.Name;
        evalOp.ProblemDefinitionParameter.ActualName = ProblemDefinitionParameter.Name;
      }
      foreach (var analyzeOp in Operators.OfType<IMultiObjectiveProgrammableProblemAnalyzer>()) {
        analyzeOp.EncodingParameter.ActualName = EncodingParameter.Name;
        analyzeOp.ProblemDefinitionParameter.ActualName = ProblemDefinitionParameter.Name;
        analyzeOp.QualitiesParameter.ActualName = Evaluator.QualitiesParameter.ActualName;
      }

      SolutionCreator = encoding.SolutionCreator;

      foreach (var param in DynamicEncodingParameters)
        if (Parameters.Contains(param)) Parameters.Remove(param);
      DynamicEncodingParameters.Clear();

      DynamicEncodingParameters.AddRange(encoding.Parameters);


      foreach (var param in DynamicEncodingParameters) {
        param.Hidden = true;
        Parameters.Add(param);
      }

    }
  }
}
